package net.sergeych.kiloparsec.adapters

import kotlinx.coroutines.*
import kotlinx.coroutines.test.runTest
import net.sergeych.crypto2.initCrypto
import net.sergeych.kiloparsec.adapter.UdpServer
import net.sergeych.kiloparsec.adapter.acceptTcpDevice
import net.sergeych.kiloparsec.adapter.connectTcpDevice
import net.sergeych.kiloparsec.adapter.toNetworkAddress
import net.sergeych.kiloparsec.decodeFromUByteArray
import net.sergeych.kiloparsec.encodeToUByteArray
import net.sergeych.mp_logger.Log
import net.sergeych.mp_logger.LogTag
import net.sergeych.synctools.ProtectedOp
import net.sergeych.synctools.invoke
import kotlin.test.Test
import kotlin.test.assertContains
import kotlin.test.assertEquals

class NetworkTest {

    @Test
    fun udpProviderTest() = runTest {
        Log.connectConsole(Log.Level.DEBUG)
        val s1 = UdpServer(17120)
        val s2 = UdpServer(17121)
        s1.send("Hello".encodeToUByteArray(), "localhost", 17121)
        val d1 = s2.incoming.receive()
        assertEquals(d1.address.port, 17120)
        assertEquals("Hello", d1.message.toByteArray().decodeToString())
        s1.send("world".encodeToUByteArray(), d1.address)
        assertEquals("world", s1.incoming.receive().message.toByteArray().decodeToString())
        //        println("s1: ${s1.bindAddress()}")

    }

    @Test
    fun tcpAsyncConnectionTest() = runTest {
        initCrypto()
        Log.connectConsole(Log.Level.DEBUG)

        coroutineScope {
            val serverFlow = acceptTcpDevice(17171)
            val op = ProtectedOp()
            var pills = setOf<String>()
            val j = launch {
                println("serf")
                serverFlow.collect { device ->
                    println("serf 0")
                    launch {
                        println("serf 1")
                        device.output.send("Hello, world!".encodeToUByteArray())
                        device.output.send("Great".encodeToUByteArray())
                        while (true) {
                            val x = device.input.receive().decodeFromUByteArray()
                            if (x.startsWith("die")) {
                                op.invoke {
                                    pills += x
                                }
                                cancel()
                            }
                            else
                                println("ignoring unexpected input: $x")
                        }
                    }
                }
            }
            yield()
            run {
                try {
                    println("pre-con")
                    val s = connectTcpDevice("127.0.1.1:17171".toNetworkAddress())
                    println("2")
                    assertEquals("Hello, world!", s.input.receive().decodeFromUByteArray())
                    assertEquals("Great", s.input.receive().decodeFromUByteArray())
                    s.output.send("Goodbye".encodeToUByteArray())
                    s.output.send("die1".encodeToUByteArray())
                    s.close()
                }
                catch(t: Throwable) {
                    t.printStackTrace()
                    throw t
                }
            }
            println("pre-con2")
            val s1 = connectTcpDevice("127.0.1.1:17171".toNetworkAddress())
            assertEquals("Hello, world!", s1.input.receive().decodeFromUByteArray())
            assertEquals("Great", s1.input.receive().decodeFromUByteArray())
            s1.output.send("die2".encodeToUByteArray())
            s1.close()

            // check that channels were flushed prior to closed:
            assertContains(pills, "die1")
            assertContains(pills, "die2")

            // Check that server jobs are closed
            j.cancelAndJoin()
        }
    }
}